import argparse
import pathlib
import random

import datafusion as dn
import datafusion.functions as dnf
import faker
import pyarrow as pa
import pyarrow.parquet as paq

import arbalister.arrow as aa
from arbalister import file_format as ff

MAX_FAKER_ROWS = 100_000


def _widen_field(field: pa.Field) -> pa.Field:
    return pa.field(field.name, pa.large_string()) if pa.types.is_string(field.type) else field


def widen(table: pa.Table) -> pa.Table:
    """Adapt Arrow schema for large files."""
    return table.cast(pa.schema([_widen_field(f) for f in table.schema]))


def generate_table(num_rows: int) -> pa.Table:
    """Generate a table with fake data."""
    if num_rows > MAX_FAKER_ROWS:
        table = widen(generate_table(MAX_FAKER_ROWS))
        n_repeat = num_rows // MAX_FAKER_ROWS
        large_table = pa.concat_tables([table] * n_repeat, promote_options="default")
        return large_table.slice(0, num_rows)

    gen = faker.Faker()
    data = {
        "name": [gen.name() for _ in range(num_rows)],
        "address": [gen.address().replace("\n", ", ") for _ in range(num_rows)],
        "age": [gen.random_number(digits=2) for _ in range(num_rows)],
        "id": [gen.uuid4() for _ in range(num_rows)],
    }
    return pa.table(data)


def _generate_coordinate_table_slice(
    row_start: int, row_end: int, num_cols: int, ctx: dn.SessionContext
) -> pa.Table:
    row_idx: pa.Array = pa.array(range(row_start, row_end), type=pa.int64())
    table = pa.table({"row": row_idx})
    table = (
        ctx.from_arrow(table)
        .with_columns(
            *[
                dnf.concat(
                    dn.lit("("),  # type: ignore[no-untyped-call]
                    dnf.col("row"),
                    dn.lit(f", {j})"),  # type: ignore[no-untyped-call]
                ).alias(f"col_{j}")
                for j in range(num_cols)
            ]
        )
        .drop("row")
        .to_arrow_table()
    )
    return widen(table)


def _sink_coordinate_table(
    num_rows: int,
    num_cols: int,
    writer: paq.ParquetWriter,
    ctx: dn.SessionContext,
    chunk_size: int = 100_000,
) -> None:
    for row_start in range(0, num_rows, chunk_size):
        print(f"Generating {row_start}", flush=True)
        row_end = min(row_start + chunk_size, num_rows)
        table = _generate_coordinate_table_slice(row_start, row_end, num_cols=num_cols, ctx=ctx)
        writer.write(table)


def sink_coordinate_table(
    num_rows: int, num_cols: int, path: pathlib.Path, chunk_size: int = 1_000_000
) -> None:
    """Write iteratively a table where each cell has its coordinates."""
    assert ff.FileFormat.from_filename(path) == ff.FileFormat.Parquet
    print("Initializing session context", flush=True)
    ctx = dn.SessionContext()
    print("Generating schema", flush=True)
    schema = _generate_coordinate_table_slice(0, 1, num_cols=num_cols, ctx=ctx).schema
    writer = paq.ParquetWriter(path, schema)
    _sink_coordinate_table(num_rows=num_rows, num_cols=num_cols, writer=writer, ctx=ctx)
    writer.close()


def configure_command_single(cmd: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Configure single subcommand CLI options."""
    cmd.add_argument("--output-file", "-o", type=pathlib.Path, required=True, help="Output file path")
    cmd.add_argument(
        "--output-type",
        "-t",
        choices=[t.name.lower() for t in ff.FileFormat],
        default=None,
        help="Output file type",
    )
    cmd.add_argument("--num-rows", type=int, default=1000, help="Number of rows to generate")
    return cmd


def configure_command_batch(cmd: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Configure batch subcommand CLI options."""
    cmd.add_argument(
        "--output-file", "-o", type=pathlib.Path, action="append", help="Output file path", default=[]
    )
    cmd.add_argument("--num-rows", type=int, default=1000, help="Number of rows to generate")
    return cmd


def configure_command_coordinate(cmd: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Configure coordinate subcommand CLI options."""
    cmd.add_argument(
        "--output-file", "-o", type=pathlib.Path, required=True, help="Output file path, must be parquet."
    )
    cmd.add_argument("--num-rows", type=int, default=1_000_000, help="Number of rows to generate")
    cmd.add_argument("--num-cols", type=int, default=1000, help="Number of rows to generate")
    return cmd


def configure_argparse() -> argparse.ArgumentParser:
    """Configure CLI options."""
    parser = argparse.ArgumentParser(description="Generate a table and write to file.")
    subparsers = parser.add_subparsers(dest="command", required=True)

    cmd_one = subparsers.add_parser("single", help="Generate a single table and write to file.")
    configure_command_single(cmd_one)

    cmd_batch = subparsers.add_parser("batch", help="Generate a multiple tables with the same data.")
    configure_command_batch(cmd_batch)

    cmd_batch = subparsers.add_parser("coordinate", help="Generate a coordinate table file.")
    configure_command_coordinate(cmd_batch)

    return parser


def shuffle_table(table: pa.Table, seed: int | None = None) -> pa.Table:
    """Shuffle the rows and columns of a table."""
    rnd = random.Random(seed)
    row_indices = pa.array(rnd.sample(range(table.num_rows), table.num_rows), type=pa.int64())
    col_order = rnd.sample(table.column_names, len(table.column_names))
    return table.select(col_order).take(row_indices)


def save_table(table: pa.Table, path: pathlib.Path, file_type: ff.FileFormat) -> None:
    """Save a table to file with the given file type."""
    path.parent.mkdir(exist_ok=True, parents=True)
    write_table = aa.get_table_writer(file_type)
    write_table(table, str(path))


def main() -> None:
    """Generate data file."""
    parser = configure_argparse()
    args = parser.parse_args()

    table = generate_table(args.num_rows)

    match args.command:
        case "single":
            ft = next((t for t in ff.FileFormat if t.name.lower() == args.output_type), None)
            if ft is None:
                ft = ff.FileFormat.from_filename(args.output_file)
            save_table(shuffle_table(table), args.output_file, ft)
        case "batch":
            for p in args.output_file:
                ft = ff.FileFormat.from_filename(p)
                save_table(shuffle_table(table), p, ft)
        case "coordinate":
            sink_coordinate_table(num_rows=args.num_rows, num_cols=args.num_cols, path=args.output_file)


if __name__ == "__main__":
    main()



import dataclasses
import pathlib
from typing import Any, Literal, Self

import adbc_driver_sqlite.dbapi as adbc_sqlite
import pyarrow as pa


def write_sqlite(
    table: pa.Table,
    path: str | pathlib.Path,
    table_name: str = "Table",
    catalog_name: str | None = None,
    mode: Literal["append", "create", "replace", "create_append"] = "create",
) -> None:
    """Write a table as an Sqlite file."""
    with adbc_sqlite.connect(str(path)) as connection:
        with connection.cursor() as cursor:
            cursor.adbc_ingest(table_name=table_name, data=table, mode=mode, catalog_name=catalog_name)
        connection.commit()


@dataclasses.dataclass
class SqliteDataFrame:
    """A DataFrame plan on a Sqlite file.

    This plan is made to resemble a Apache Datafusion DataFrame despite some performances penalty.
    This is because in the Rust Datafusion implementation of as a Sqlite contrib which we would
    like to eventually use so this class is only filling in temporarily.
    """

    _path: str
    _table_name: str
    _schema: pa.Schema
    _num_rows: int
    _limit: int | None = None
    _offset: int | None = None
    _select: list[str] | None = None

    @staticmethod
    def get_table_names(path: pathlib.Path | str) -> list[str]:
        """Get the list of table names in a SQLite database."""
        with adbc_sqlite.connect(str(path)) as connection:
            with connection.cursor() as cursor:
                cursor.execute("SELECT name FROM sqlite_master WHERE type='table'")
                return [row for (row,) in cursor.fetchall()]

    @classmethod
    def read_sqlite(cls, context: Any, path: pathlib.Path | str, table_name: str | None = None) -> Self:
        """Read an Sqlite file metadata and start a new DataFrame plan."""
        tables = cls.get_table_names(path)

        if table_name is None and len(tables) > 0:
            table_name = tables[0]
        if table_name not in tables or table_name is None:
            raise ValueError(f"Invalid table name {table_name}")

        with adbc_sqlite.connect(str(path)) as connection:
            with connection.cursor() as cursor:
                cursor.execute(f'SELECT COUNT(*) FROM "{table_name}"')
                num_rows = cursor.fetchone()[0]  # type: ignore[index]

            schema = connection.adbc_get_table_schema(table_name)

            return cls(_path=str(path), _table_name=table_name, _schema=schema, _num_rows=num_rows)

    def schema(self) -> pa.Schema:
        """Return the :py:class:`pyarrow.Schema` of this DataFrame."""
        return self._schema

    def limit(self, count: int, offset: int = 0) -> Self:
        """Return a new DataFrame with a limited number of rows."""
        return dataclasses.replace(self, _limit=count, _offset=offset)

    def select(self, *columns: str) -> Self:
        """Project arbitrary expressions into a new `DataFrame`."""
        return dataclasses.replace(self, _select=list(columns))

    def count(self) -> int:
        """Return the total number of rows in this DataFrame."""
        return self._num_rows

    def to_arrow_table(self) -> pa.Table:
        """Execute the DataFrame and convert it into an Arrow Table."""
        limit = (
            f"LIMIT {self._limit} OFFSET {self._offset}"
            if self._limit is not None and self._offset is not None
            else ""
        )
        columns = self._select if self._select is not None else ["*"]

        with adbc_sqlite.connect(self._path) as connection:
            with connection.cursor() as cursor:
                cursor.execute(f'SELECT {",".join(columns)} FROM "{self._table_name}" {limit}')
                return cursor.fetch_arrow_table()



import enum
import pathlib
from typing import Self


class FileFormat(enum.StrEnum):
    """Known file format that we can read into an Arrow format.

    Todo:
    - ADBC (Sqlite/Postgres)

    """

    Avro = "avro"
    Csv = "csv"
    Ipc = "ipc"
    Orc = "orc"
    Parquet = "parquet"
    Sqlite = "sqlite"

    @classmethod
    def from_filename(cls, file: pathlib.Path | str) -> Self:
        """Get the file format from a filename extension."""
        file_type = pathlib.Path(file).suffix.removeprefix(".").strip().lower()

        # Match again their default value
        if ft := next((ft for ft in FileFormat if str(ft) == file_type), None):
            return ft
        # Match other known values
        match file_type:
            case "ipc" | "feather" | "arrow":
                return cls.Ipc
            case "sqlite3" | "db" | "db3" | "s3db" | "sl3":
                return cls.Sqlite
        raise ValueError(f"Unknown file type {file_type}")



import dataclasses
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import arbalister.params as params


@dataclasses.dataclass
class SimpleExample:
    """A simple test dataclass."""

    count: int
    label: str
    flag: bool


@dataclasses.dataclass
class OptionalExample:
    """A test dataclass with optional."""

    value: int | None
    text: Optional[str]
    flag: Optional[bool]


@dataclasses.dataclass(frozen=True, slots=True)
class UnionExample:
    """A test dataclass with union."""

    data: Union[int, str]
    number_or_float: int | float


@dataclasses.dataclass
class DefaultExample:
    """A test dataclass with default."""

    count: int = 3
    name: str = "default"
    enabled: bool = True
    generated: str = dataclasses.field(default_factory=lambda: "factory")


def _callback_from_mapping(responses: Dict[str, Any]) -> Callable[[str, Any], Any]:
    def _cb(name: str, default: Any) -> Any:
        return responses.get(name, default)

    return _cb


def test_simple_fields() -> None:
    """Simple builder."""
    responses: Dict[str, Any] = {"count": "42", "label": 99, "flag": "yes"}
    result = params.build_dataclass(SimpleExample, _callback_from_mapping(responses))
    assert result == SimpleExample(count=42, label="99", flag=True)


def test_optional_fields_with_none() -> None:
    """None parsed as optional."""
    responses: Dict[str, Any] = {"value": "None", "text": None, "flag": "off"}
    result = params.build_dataclass(OptionalExample, _callback_from_mapping(responses))
    assert result == OptionalExample(value=None, text=None, flag=False)


def test_optional_fields_with_values() -> None:
    """Values parsed as optional."""
    responses: Dict[str, Any] = {"value": "7", "text": "hello", "flag": True}
    result = params.build_dataclass(OptionalExample, _callback_from_mapping(responses))
    assert result == OptionalExample(value=7, text="hello", flag=True)


def test_union_fields_int_branch() -> None:
    """Union are parsed."""
    responses: Dict[str, Any] = {"data": "21", "number_or_float": "8"}
    result = params.build_dataclass(UnionExample, _callback_from_mapping(responses))
    assert result == UnionExample(data=21, number_or_float=8)


def test_union_fields_str_and_float_branch() -> None:
    """Union are parsed."""
    responses: Dict[str, Any] = {"data": "abc", "number_or_float": "3.14"}
    result = params.build_dataclass(UnionExample, _callback_from_mapping(responses))
    assert result == UnionExample(data="abc", number_or_float=3.14)


def test_defaults_passed_to_callback() -> None:
    """Default are used in absence of value."""
    seen: List[Tuple[str, Any]] = []

    def _cb(name: str, default: Any) -> Any:
        seen.append((name, default))
        return default

    result = params.build_dataclass(DefaultExample, _cb)
    assert result == DefaultExample()
    assert seen == [
        ("count", 3),
        ("name", "default"),
        ("enabled", True),
        ("generated", "factory"),
    ]


def test_overriding_defaults() -> None:
    """Default are ignored in when a value is provided."""
    responses: Dict[str, Any] = {"count": "10", "name": "custom", "generated": "supplied"}
    result = params.build_dataclass(DefaultExample, _callback_from_mapping(responses))
    assert result == DefaultExample(count=10, name="custom", enabled=True, generated="supplied")


def test_bool_parsing_variants() -> None:
    """Truthy values are parsed as bool."""
    responses: Dict[str, Any] = {"count": "1", "label": "x", "flag": "FALSE"}
    result = params.build_dataclass(SimpleExample, _callback_from_mapping(responses))
    assert result == SimpleExample(count=1, label="x", flag=False)


def test_float_parsing_in_union() -> None:
    """Float is a better match than int in union."""
    responses: Dict[str, Any] = {"data": "5.0", "number_or_float": 2.5}
    result = params.build_dataclass(UnionExample, _callback_from_mapping(responses))
    assert result == UnionExample(data="5.0", number_or_float=2.5)


@dataclasses.dataclass
class Address:
    """A nested dataclass."""

    street: str
    city: str
    zip_code: int


@dataclasses.dataclass
class Person:
    """A dataclass with nested dataclass field."""

    name: str
    age: int
    address: Address


def test_nested_dataclass_from_dict() -> None:
    """Nested dataclass parsed from dict."""
    responses: Dict[str, Any] = {
        "name": "Alice",
        "age": "30",
        "address": {"street": "123 Main St", "city": "Springfield", "zip_code": "12345"},
    }
    result = params.build_dataclass(Person, _callback_from_mapping(responses))
    assert result == Person(
        name="Alice", age=30, address=Address(street="123 Main St", city="Springfield", zip_code=12345)
    )


def test_nested_dataclass_from_json_string() -> None:
    """Nested dataclass parsed from JSON string."""
    responses: Dict[str, Any] = {
        "name": "Bob",
        "age": "25",
        "address": '{"street": "456 Oak Ave", "city": "Shelbyville", "zip_code": "67890"}',
    }
    result = params.build_dataclass(Person, _callback_from_mapping(responses))
    assert result == Person(
        name="Bob", age=25, address=Address(street="456 Oak Ave", city="Shelbyville", zip_code=67890)
    )


def test_nested_dataclass_already_instance() -> None:
    """Nested dataclass already an instance."""
    addr = Address(street="789 Pine Rd", city="Capital City", zip_code=11111)
    responses: Dict[str, Any] = {"name": "Charlie", "age": "35", "address": addr}
    result = params.build_dataclass(Person, _callback_from_mapping(responses))
    assert result == Person(name="Charlie", age=35, address=addr)



"""Jupyter server extension to serve content in Arrow format."""

try:
    from ._version import __version__
except ImportError:
    # Fallback when using the package in dev mode without installing
    # in editable mode with pip. It is highly recommended to install
    # the package from a stable release or in editable mode:
    # https://pip.pypa.io/en/stable/topics/local-project-installs/#editable-installs
    import warnings

    warnings.warn("Importing 'arbalister' outside a proper installation.", stacklevel=1)
    __version__ = "dev"

import jupyterlab.labapp

from . import arrow as arrow
from . import file_format as file_format
from . import params as params
from . import routes as routes


def _jupyter_labextension_paths() -> list[dict[str, str]]:
    return [{"src": "labextension", "dest": "arbalister"}]


def _jupyter_server_extension_points() -> list[dict[str, str]]:
    return [{"module": "arbalister"}]


def _load_jupyter_server_extension(server_app: jupyterlab.labapp.LabApp) -> None:
    """Register the API handler to receive HTTP requests from the frontend extension."""
    routes.setup_route_handlers(server_app.web_app)
    name = "arbalister"
    server_app.log.info(f"Registered {name} server extension")



from typing import Any

import pytest

pytest_plugins = ("pytest_jupyter.jupyter_server",)


@pytest.fixture
def jp_server_config(jp_server_config: Any) -> dict[str, Any]:
    """Set up jupyter server configuration."""
    return {"ServerApp": {"jpserver_extensions": {"arbalister": True}}}



import base64
import dataclasses
import os
import pathlib

import datafusion as dn
import datafusion.functions as dnf
import jupyter_server.base.handlers
import jupyter_server.serverapp
import pyarrow as pa
import tornado
from jupyter_server.utils import url_path_join

from . import arrow as abw
from . import file_format as ff
from . import params as params


@dataclasses.dataclass(frozen=True, slots=True)
class Empty:
    """An empty data class."""


@dataclasses.dataclass(frozen=True, slots=True)
class SqliteReadOptions:
    """Query parameter for the Sqlite reader."""

    table_name: str | None = None


@dataclasses.dataclass(frozen=True, slots=True)
class CsvReadOptions:
    """Query parameter for the CSV reader."""

    delimiter: str | None = ","


FileReadOptions = SqliteReadOptions | CsvReadOptions | Empty


class BaseRouteHandler(jupyter_server.base.handlers.APIHandler):
    """A base handler to share common methods."""

    def initialize(self, context: dn.SessionContext) -> None:
        """Process custom constructor arguments."""
        super().initialize()
        self.context = context

    def data_file(self, path: str) -> pathlib.Path:
        """Return the file that is requested by the URL path."""
        root_dir = pathlib.Path(os.path.expanduser(self.settings["server_root_dir"])).resolve()
        return root_dir / path

    def dataframe(self, path: str) -> dn.DataFrame:
        """Return the DataFusion lazy DataFrame.

        Note: On some file type, the file is read eagerly when calling this method.
        """
        file = self.data_file(path)
        file_format = ff.FileFormat.from_filename(file)
        file_params = self.get_file_options(file_format)
        read_table = abw.get_table_reader(format=file_format)
        return read_table(self.context, file, **dataclasses.asdict(file_params))

    def get_query_params_as[T](self, dataclass_type: type[T]) -> T:
        """Extract query parameters into a dataclass type."""
        return params.build_dataclass(dataclass_type, self.get_query_argument)

    def get_file_options(self, file_format: ff.FileFormat) -> FileReadOptions:
        """Read the parameters associated with the relevant file format."""
        match file_format:
            case ff.FileFormat.Sqlite:
                return self.get_query_params_as(SqliteReadOptions)
            case ff.FileFormat.Csv:
                return self.get_query_params_as(CsvReadOptions)
        return Empty()


@dataclasses.dataclass(frozen=True, slots=True)
class IpcParams:
    """Query parameter for IPC data."""

    start_row: int | None = None
    end_row: int | None = None
    start_col: int | None = None
    end_col: int | None = None


class IpcRouteHandler(BaseRouteHandler):
    """An handler to get file in IPC."""

    @tornado.web.authenticated
    async def get(self, path: str) -> None:
        """HTTP GET return an IPC file."""
        params = self.get_query_params_as(IpcParams)

        self.set_header("Content-Type", "application/vnd.apache.arrow.stream")

        df: dn.DataFrame = self.dataframe(path)

        if params.start_row is not None and params.end_row is not None:
            offset: int = params.start_row
            count: int = params.end_row - params.start_row
            df = df.limit(count=count, offset=offset)

        if params.start_col is not None and params.end_col is not None:
            col_names = df.schema().names
            df = df.select(*col_names[params.start_col : params.end_col])

        table: pa.Table = df.to_arrow_table()

        # TODO can we write directly to socket and send chunks
        sink = pa.BufferOutputStream()
        with pa.ipc.new_stream(sink, table.schema) as writer:
            writer.write_table(table)

        buf: pa.Buffer = sink.getvalue()
        self.write(buf.to_pybytes())  # FIXME to_pybytes copies memory

        await self.flush()


@dataclasses.dataclass(frozen=True, slots=True)
class SchemaInfo:
    """Schema information as a zero-row IPC stream."""

    data: str
    mimetype: str = "application/vnd.apache.arrow.stream"
    encoding: str = "base64"


@dataclasses.dataclass(frozen=True, slots=True)
class StatsResponse:
    """File statistics returned in the stats route."""

    schema: SchemaInfo
    num_rows: int = 0
    num_cols: int = 0


class StatsRouteHandler(BaseRouteHandler):
    """An handler to get file in IPC."""

    @tornado.web.authenticated
    async def get(self, path: str) -> None:
        """HTTP GET return statistics."""
        df = self.dataframe(path)

        # FIXME this is not optimal for ORC/CSV where we can read_metadata, but it is not read
        # via DataFusion.
        schema = df.schema()
        try:
            num_rows = df.count()
        # Workaround issue in Avro files df.count() not working
        except Exception as e:
            if len(schema.names) == 0:
                num_rows = 0
            # No dedicated exception type coming from DataFusion
            if str(e).startswith("DataFusion"):
                first_col: str = schema.names[0]
                batches = df.aggregate([], [dnf.count(dn.col(first_col))]).collect()
                num_rows = batches[0].column(0)[0].as_py()

        # Create a zero-row IPC stream with the table schema
        zero_row_table = df.limit(0, 0).to_arrow_table()

        sink = pa.BufferOutputStream()
        with pa.ipc.new_stream(sink, schema) as writer:
            writer.write_table(zero_row_table)

        buf: pa.Buffer = sink.getvalue()
        schema_64 = base64.b64encode(buf.to_pybytes()).decode("utf-8")

        response = StatsResponse(
            num_cols=len(schema),
            num_rows=num_rows,
            schema=SchemaInfo(data=schema_64),
        )
        await self.finish(dataclasses.asdict(response))


@dataclasses.dataclass(frozen=True, slots=True)
class SqliteFileInfo:
    """Sqlite specific information about a file."""

    table_names: list[str]


@dataclasses.dataclass(frozen=True, slots=True)
class CsvFileInfo:
    """Csv specific information about a file."""

    delimiters: list[str] = dataclasses.field(default_factory=lambda: [",", ";", "\\t", "|", "#"])


FileInfo = SqliteFileInfo | CsvFileInfo


@dataclasses.dataclass(frozen=True, slots=True)
class FileInfoResponse[I, P]:
    """File-specific information and defaults returned in the file info route."""

    info: I
    default_options: P


CsvFileInfoResponse = FileInfoResponse[CsvFileInfo, CsvReadOptions]
SqliteFileInfoResponse = FileInfoResponse[SqliteFileInfo, SqliteReadOptions]

NoFileInfoResponse = FileInfoResponse[Empty, Empty]


class FileInfoRouteHandler(BaseRouteHandler):
    """A handler to get file-specific information."""

    @tornado.web.authenticated
    async def get(self, path: str) -> None:
        """HTTP GET return file-specific information."""
        file = self.data_file(path)
        file_format = ff.FileFormat.from_filename(file)

        match file_format:
            case ff.FileFormat.Csv:
                info = CsvFileInfo()
                csv_response = CsvFileInfoResponse(
                    info=info,
                    default_options=CsvReadOptions(delimiter=info.delimiters[0]),
                )
                await self.finish(dataclasses.asdict(csv_response))
            case ff.FileFormat.Sqlite:
                from . import adbc

                table_names = adbc.SqliteDataFrame.get_table_names(file)

                sqlite_response = SqliteFileInfoResponse(
                    info=SqliteFileInfo(table_names=table_names),
                    default_options=SqliteReadOptions(table_name=table_names[0]),
                )
                await self.finish(dataclasses.asdict(sqlite_response))
            case _:
                no_response = NoFileInfoResponse(info=Empty(), default_options=Empty())
                await self.finish(dataclasses.asdict(no_response))


def make_datafusion_config() -> dn.SessionConfig:
    """Return the datafusion config."""
    config = (
        dn.SessionConfig()
        # Must use a single partition otherwise limit parallelism will return arbitrary rows
        .with_target_partitions(1)
        # String views do not get written properly to IPC
        .set("datafusion.execution.parquet.schema_force_view_types", "false")
    )

    return config


def setup_route_handlers(web_app: jupyter_server.serverapp.ServerWebApplication) -> None:
    """Jupyter server setup entry point."""
    host_pattern = ".*$"
    base_url = web_app.settings["base_url"]

    context = dn.SessionContext(make_datafusion_config())

    handlers = [
        (url_path_join(base_url, r"arrow/stream/([^?]*)"), IpcRouteHandler, {"context": context}),
        (url_path_join(base_url, r"arrow/stats/([^?]*)"), StatsRouteHandler, {"context": context}),
        (url_path_join(base_url, r"file/info/([^?]*)"), FileInfoRouteHandler, {"context": context}),
    ]

    web_app.add_handlers(host_pattern, handlers)  # type: ignore[no-untyped-call]



import base64
import dataclasses
import json
import pathlib
import random
import string
from typing import Awaitable, Callable

import pyarrow as pa
import pytest
import tornado

import arbalister as arb
import arbalister.file_format as ff


@pytest.fixture(
    params=[
        (ff.FileFormat.Avro, arb.routes.Empty()),
        (ff.FileFormat.Csv, arb.routes.Empty()),
        (ff.FileFormat.Csv, arb.routes.CsvReadOptions(delimiter=";")),
        (ff.FileFormat.Ipc, arb.routes.Empty()),
        (ff.FileFormat.Orc, arb.routes.Empty()),
        (ff.FileFormat.Parquet, arb.routes.Empty()),
        (ff.FileFormat.Sqlite, arb.routes.Empty()),
        (ff.FileFormat.Sqlite, arb.routes.SqliteReadOptions(table_name="dummy_table_2")),
    ],
    ids=lambda f_p: f"{f_p[0].value}-{dataclasses.asdict(f_p[1])}",
    scope="module",
)
def file_format_and_params(
    request: pytest.FixtureRequest,
) -> tuple[ff.FileFormat, arb.routes.FileReadOptions]:
    """Parametrize the file format and file parameters used in the tests.

    This is used to to build test cases with a give set of parameters since each file format may be tested
    with a different number of parameters.
    """
    out: tuple[ff.FileFormat, arb.routes.FileReadOptions] = request.param
    return out


@pytest.fixture(scope="module")
def file_format(file_format_and_params: tuple[ff.FileFormat, arb.routes.FileReadOptions]) -> ff.FileFormat:
    """Extract the the file format fixture value used in the tests."""
    return file_format_and_params[0]


@pytest.fixture(scope="module")
def file_params(
    file_format_and_params: tuple[ff.FileFormat, arb.routes.FileReadOptions],
) -> arb.routes.FileReadOptions:
    """Extract the the file parameters fixture value used in the tests."""
    return file_format_and_params[1]


@pytest.fixture(scope="module")
def dummy_table_1(num_rows: int = 10) -> pa.Table:
    """Generate a table with fake data."""
    data = {
        "lower": random.choices(string.ascii_lowercase, k=num_rows),
        "sequence": list(range(num_rows)),
        "upper": random.choices(string.ascii_uppercase, k=num_rows),
        "number": [random.random() for _ in range(num_rows)],
    }
    table = pa.table(data)
    return table


@pytest.fixture(scope="module")
def dummy_table_2(num_rows: int = 13) -> pa.Table:
    """Generate a table with different fake data."""
    data = {
        "id": list(range(num_rows)),
        "flag": [random.choice([True, False]) for _ in range(num_rows)],
        "letter": random.choices(string.ascii_letters, k=num_rows),
        "score": [random.randint(0, 100) for _ in range(num_rows)],
        "timestamp": [random.randint(1_600_000_000, 1_700_000_000) for _ in range(num_rows)],
    }
    table = pa.table(data)
    return table


@pytest.fixture(scope="module")
def full_table(file_params: ff.FileFormat, dummy_table_1: pa.Table, dummy_table_2: pa.Table) -> pa.Table:
    """Return the full table on which we are executed queries."""
    if isinstance(file_params, arb.routes.SqliteReadOptions):
        return {
            "dummy_table_1": dummy_table_1,
            "dummy_table_2": dummy_table_2,
        }[file_params.table_name]
    return dummy_table_1


@pytest.fixture
def table_file(
    jp_root_dir: pathlib.Path,
    dummy_table_1: pa.Table,
    dummy_table_2: pa.Table,
    file_format: ff.FileFormat,
    file_params: arb.routes.FileReadOptions,
) -> pathlib.Path:
    """Write the dummy table to file."""
    write_table = arb.arrow.get_table_writer(file_format)
    table_path = jp_root_dir / f"test.{str(file_format).lower()}"

    match file_format:
        case ff.FileFormat.Csv:
            write_table(dummy_table_1, table_path, delimiter=getattr(file_params, "delimiter", ","))
        case ff.FileFormat.Sqlite:
            write_table(dummy_table_1, table_path, table_name="dummy_table_1", mode="create_append")
            write_table(dummy_table_2, table_path, table_name="dummy_table_2", mode="create_append")
        case _:
            write_table(dummy_table_1, table_path)

    return table_path.relative_to(jp_root_dir)


JpFetch = Callable[..., Awaitable[tornado.httpclient.HTTPResponse]]


@pytest.fixture(
    params=[
        # No limits
        lambda table: arb.routes.IpcParams(),
        # Limit only number of rows
        lambda table: arb.routes.IpcParams(start_row=0, end_row=3),
        lambda table: arb.routes.IpcParams(start_row=2, end_row=4),
        lambda table: arb.routes.IpcParams(start_row=0, end_row=table.num_rows),
        lambda table: arb.routes.IpcParams(start_row=table.num_rows // 2, end_row=table.num_rows),
        # Limit only number of cols
        lambda table: arb.routes.IpcParams(start_col=0, end_col=3),
        lambda table: arb.routes.IpcParams(start_col=2, end_col=4),
        lambda table: arb.routes.IpcParams(start_col=0, end_col=table.num_columns),
        lambda table: arb.routes.IpcParams(start_col=table.num_columns // 2, end_col=table.num_columns),
        # Limit both
        lambda table: arb.routes.IpcParams(
            start_row=0,
            end_row=3,
            start_col=table.num_columns // 2,
            end_col=table.num_columns,
        ),
        lambda table: arb.routes.IpcParams(
            start_row=0,
            end_row=table.num_rows,
            start_col=2,
            end_col=4,
        ),
        # Schema only
        lambda table: arb.routes.IpcParams(
            start_row=0,
            end_row=0,
        ),
    ]
)
def ipc_params(request: pytest.FixtureRequest, dummy_table_1: pa.Table) -> arb.routes.IpcParams:
    """Parameters used to select the IPC data in the response."""
    make_table: Callable[[pa.Table], arb.routes.IpcParams] = request.param
    return make_table(dummy_table_1)


async def test_ipc_route_limit(
    jp_fetch: JpFetch,
    full_table: pa.Table,
    table_file: pathlib.Path,
    ipc_params: arb.routes.IpcParams,
    file_params: arb.routes.FileReadOptions,
    file_format: ff.FileFormat,
) -> None:
    """Test fetching a file returns the limited rows and columns in IPC."""
    response = await jp_fetch(
        "arrow/stream",
        str(table_file),
        params={
            k: v
            for k, v in {**dataclasses.asdict(ipc_params), **dataclasses.asdict(file_params)}.items()
            if v is not None
        },
    )

    assert response.code == 200
    assert response.headers["Content-Type"] == "application/vnd.apache.arrow.stream"
    payload = pa.ipc.open_stream(response.body).read_all()

    expected = full_table

    # Row slicing
    if (start_row := ipc_params.start_row) is not None and (end_row := ipc_params.end_row) is not None:
        expected_num_rows = min(end_row, expected.num_rows) - start_row
        assert payload.num_rows == expected_num_rows
        expected = expected.slice(start_row, end_row - start_row)

    # Col slicing
    if (start_col := ipc_params.start_col) is not None and (end_col := ipc_params.end_col) is not None:
        expected_num_cols = min(end_col, len(expected.schema)) - start_col
        assert len(payload.schema) == expected_num_cols
        col_names = expected.schema.names
        expected = expected.select(col_names[start_col:end_col])

    assert expected.cast(payload.schema) == payload


async def test_stats_route(
    jp_fetch: JpFetch,
    full_table: pa.Table,
    table_file: pathlib.Path,
    file_params: arb.routes.FileReadOptions,
    file_format: ff.FileFormat,
) -> None:
    """Test fetching a file returns the correct metadata in Json."""
    response = await jp_fetch(
        "arrow/stats/",
        str(table_file),
        params={k: v for k, v in dataclasses.asdict(file_params).items() if v is not None},
    )

    assert response.code == 200
    assert response.headers["Content-Type"] == "application/json; charset=UTF-8"

    payload = json.loads(response.body)

    assert payload["num_cols"] == len(full_table.schema)
    assert payload["num_rows"] == full_table.num_rows
    assert payload["schema"]["mimetype"] == "application/vnd.apache.arrow.stream"
    assert payload["schema"]["encoding"] == "base64"
    table_64 = base64.b64decode(payload["schema"]["data"])
    table = pa.ipc.open_stream(table_64).read_all()
    assert table.num_rows == 0
    assert table.schema.names == full_table.schema.names


async def test_file_info_route_sqlite(
    jp_fetch: JpFetch,
    table_file: pathlib.Path,
    file_format: ff.FileFormat,
) -> None:
    """Test fetching file info for SQLite files returns table names."""
    response = await jp_fetch("file/info/", str(table_file))

    assert response.code == 200
    assert response.headers["Content-Type"] == "application/json; charset=UTF-8"

    payload = json.loads(response.body)
    info = payload["info"]
    default_options = payload["default_options"]

    match file_format:
        case ff.FileFormat.Csv:
            assert isinstance(info["delimiters"], list)
            assert "," in info["delimiters"]
            assert default_options["delimiter"] == info["delimiters"][0]
        case ff.FileFormat.Sqlite:
            assert isinstance(info["table_names"], list)
            assert "dummy_table_1" in info["table_names"]
            assert "dummy_table_2" in info["table_names"]
            assert default_options["table_name"] == info["table_names"][0]



import codecs
import pathlib
from typing import Any, Callable

import datafusion as dn
import pyarrow as pa

from . import file_format as ff

ReadCallable = Callable[..., dn.DataFrame]


def _read_csv(
    ctx: dn.SessionContext, path: str | pathlib.Path, delimiter: str, **kwargs: dict[str, Any]
) -> dn.DataFrame:
    if len(delimiter) > 1:
        delimiter = codecs.decode(delimiter, "unicode_escape")
    return ctx.read_csv(path, delimiter=delimiter, **kwargs)  # type: ignore[arg-type]


def _read_ipc(ctx: dn.SessionContext, path: str | pathlib.Path, **kwargs: dict[str, Any]) -> dn.DataFrame:
    import pyarrow.feather

    #  table = pyarrow.feather.read_table(path, {**{"memory_map": True}, **kwargs})
    table = pyarrow.feather.read_table(path, **kwargs)
    return ctx.from_arrow(table)


def _read_orc(ctx: dn.SessionContext, path: str | pathlib.Path, **kwargs: dict[str, Any]) -> dn.DataFrame:
    # Watch for https://github.com/datafusion-contrib/datafusion-orc
    # Evolution for native datafusion reader
    import pyarrow.orc

    table = pyarrow.orc.read_table(path, **kwargs)
    return ctx.from_arrow(table)


def get_table_reader(format: ff.FileFormat) -> ReadCallable:
    """Get the datafusion reader factory function for the given format."""
    # TODO: datafusion >= 50.0
    #  def read(ctx: dtfn.SessionContext, path: str | pathlib.Path, *args, **kwargs) -> dtfn.DataFrame:
    #      ds = pads.dataset(source=path, format=format.value)
    #      return ctx.read_table(ds, *args, **kwargs)
    out: ReadCallable
    match format:
        case ff.FileFormat.Avro:
            out = dn.SessionContext.read_avro
        case ff.FileFormat.Csv:
            out = _read_csv
        case ff.FileFormat.Parquet:
            out = dn.SessionContext.read_parquet
        case ff.FileFormat.Ipc:
            out = _read_ipc
        case ff.FileFormat.Orc:
            out = _read_orc
        case ff.FileFormat.Sqlite:
            from . import adbc as adbc

            # FIXME: For now we just pretend SqliteDataFrame is a datafusion DataFrame
            # Either we integrate it properly into Datafusion, or we create a DataFrame as a
            # typing.protocol.
            out = adbc.SqliteDataFrame.read_sqlite  # type: ignore[assignment]

    return out


WriteCallable = Callable[..., None]


def _arrow_to_avro_type(field: pa.Field) -> str | dict[str, Any]:
    t = field.type
    if pa.types.is_integer(t):
        return "long" if t.bit_width > 32 else "int"
    if pa.types.is_floating(t):
        return "double" if t.bit_width > 32 else "float"
    if pa.types.is_boolean(t):
        return "boolean"
    if pa.types.is_string(t):
        return "string"
    if pa.types.is_binary(t) or pa.types.is_large_binary(t):
        return "bytes"
    if pa.types.is_timestamp(t):
        return "long"
    if pa.types.is_list(t) or pa.types.is_large_list(t):
        item = _arrow_to_avro_type(pa.field("item", t.value_type))
        return {"type": "array", "items": item}
    # fallback
    return "string"


def _write_avro(
    table: pa.Table, path: str | pathlib.Path, name: str = "Record", namespace: str = "ns"
) -> None:
    # Avro writing is an optional dependency not added by default as it is only necessary during testing
    import json

    import avro.schema
    from avro.datafile import DataFileWriter
    from avro.io import DatumWriter

    schema = {
        "type": "record",
        "name": name,
        "namespace": namespace,
        "fields": [{"name": f.name, "type": _arrow_to_avro_type(f)} for f in table.schema],
    }
    schema_parsed = avro.schema.parse(json.dumps(schema))
    recs = table.to_pylist()
    with open(path, "wb") as f:
        writer = DataFileWriter(f, DatumWriter(), schema_parsed)
        for rec in recs:
            writer.append(rec)
        writer.close()


def get_table_writer(format: ff.FileFormat) -> WriteCallable:
    """Get the arrow writer factory function for the given format."""
    out: WriteCallable
    match format:
        case ff.FileFormat.Avro:
            out = _write_avro
        case ff.FileFormat.Csv:
            import pyarrow.csv

            def write_csv(
                data: pa.Table,
                output_file: str | pathlib.Path,
                memory_pool: pa.MemoryPool | None = None,
                **kwargs: dict[str, Any],
            ) -> None:
                pyarrow.csv.write_csv(
                    data=data,
                    output_file=str(output_file),
                    memory_pool=memory_pool,
                    write_options=pyarrow.csv.WriteOptions(**kwargs),
                )

            out = write_csv
        case ff.FileFormat.Parquet:
            import pyarrow.parquet

            out = pyarrow.parquet.write_table
        case ff.FileFormat.Ipc:
            import pyarrow.feather

            out = pyarrow.feather.write_feather
        case ff.FileFormat.Orc:
            import pyarrow.orc

            out = pyarrow.orc.write_table
        case ff.FileFormat.Sqlite:
            from . import adbc as adbc

            out = adbc.write_sqlite
    return out



import dataclasses
from types import NoneType, UnionType
from typing import Any, Callable, Union, cast, get_args, get_origin


def _parse_bool(value: Any) -> bool:
    match value:
        case bool():
            return value
        case str():
            lower = value.strip().lower()
            if lower in {"true", "1", "yes", "y", "on"}:
                return True
            if lower in {"false", "0", "no", "n", "off"}:
                return False
        case int() | float():
            return bool(value)
    raise ValueError(f"Cannot parse bool from {value!r}")


def _convert(value: Any, target_type: type[Any]) -> Any:
    match target_type:
        case _ if target_type is bool:
            return _parse_bool(value)
        case _ if target_type is int:
            return int(value)
        case _ if target_type is float:
            return float(value)
        case _ if target_type is str:
            return str(value)
        case _:
            raise TypeError(f"Unsupported type {target_type!r}")


def _ordered_union_args(value: Any, args: tuple[type[Any], ...]) -> list[type[Any]]:
    non_none = [arg for arg in args if arg is not NoneType]
    if {int, float}.issuperset(set(non_none)) and len(non_none) == 2:
        if isinstance(value, float) and not value.is_integer():
            return [float, int]
        if isinstance(value, str) and any(ch in value.lower() for ch in (".", "e")):
            return [float, int]
    return non_none


def _parse_dataclass(value: Any, dataclass_type: type[Any]) -> Any:
    """Parse a value into a dataclass instance."""
    # If already an instance of the dataclass, return as-is
    if isinstance(value, dataclass_type):
        return value

    # If it's a dict, build the dataclass from it
    if isinstance(value, dict):
        return build_dataclass(dataclass_type, lambda name, default: value.get(name, default))

    # If it's a string, try to parse as JSON first
    if isinstance(value, str):
        import json

        try:
            parsed = json.loads(value)
            if isinstance(parsed, dict):
                return build_dataclass(dataclass_type, lambda name, default: parsed.get(name, default))
        except (json.JSONDecodeError, ValueError):
            pass

    raise TypeError(f"Cannot convert {type(value).__name__} to {dataclass_type.__name__}")


def _parse_union(value: Any, args: tuple[type[Any], ...]) -> Any:
    has_none = any(arg is NoneType for arg in args)
    if value is None or (isinstance(value, str) and value.strip().lower() == "none"):
        if has_none:
            return None
    last_error: Exception | None = None
    for arg in _ordered_union_args(value, args):
        try:
            return _parse_value(value, arg)
        except Exception as exc:  # noqa: BLE001
            last_error = exc
    if last_error is not None:
        raise last_error
    raise TypeError("No matching union type")


def _parse_value(value: Any, annotation: Any) -> Any:
    origin = get_origin(annotation)
    match origin:
        case _ if origin is Union or origin is UnionType:
            return _parse_union(value, get_args(annotation))
        case None:
            # Check if annotation is a dataclass type (not an instance)
            if dataclasses.is_dataclass(annotation) and isinstance(annotation, type):
                return _parse_dataclass(value, annotation)
            return _convert(value, annotation)
        case _:
            return _convert(value, origin)


def build_dataclass[T](dataclass_type: type[T], callback: Callable[[str, Any], Any]) -> T:
    """Build a dtaclass from its definition and a value-p[rovising callback."""
    values: dict[str, Any] = {}
    for field in dataclasses.fields(cast(type[Any], dataclass_type)):
        default = (
            field.default
            if field.default is not dataclasses.MISSING
            else (
                field.default_factory()
                if field.default_factory is not dataclasses.MISSING
                else dataclasses.MISSING
            )
        )
        raw = callback(field.name, default)
        values[field.name] = _parse_value(raw, field.type)
    return dataclass_type(**values)
